@@ -124,6 +124,19 @@ class Args:
                 self.keywords.append(value)
 
 
+class Parser(argparse.ArgumentParser):
+    """
+    Customised argparse class with custom error messaging
+    """
+    def error(self, message: str) -> None:
+        """
+        Override the error method to output help instead of usage
+        """
+        self.print_help()  # Display the help info instead of usage
+        log.critical(f'\nERROR: {message}')  # Add the error at the bottom
+        sys.exit(2)  # Exit with an error code
+
+
 # Map to enhance CLI argument parser by providing additional parsing
 #     context extracted from parameter annotation data types.
 ANNOTATION_MAP = {
@@ -164,6 +177,110 @@ log = logging.getLogger('testing')
 log.setLevel(logging.CRITICAL)
 
 
+def _reconcile_args(parsed_args: argparse.Namespace, filePath: str) -> None:
+    """
+    Process arguments from a static config file against arguments provided
+        at the CLI. Prefer CLI-provided arguments over file-based ones. Also
+        prefer the MERAKI_DASHBOARD_API_KEY environment variable over one
+        provided by a config file.
+
+    - parsed_args: Parsed arguments namespace object from argparse.
+    - filePath: String file path for an existing config file
+    """
+    if parsed_args.debug:
+        log.warning(f'INFO: Loading config file: {filePath}')
+    try:  # Catch a JSON parsing failure
+        # Open file and load contents with JSON
+        fileData = json.loads(open(filePath).read())
+    except json.decoder.JSONDecodeError as e:
+        # If there is bad JSON data, throw error and exit
+        log.critical(f'ERROR: Config file {filePath} contains malformatted'
+                     ' JSON data')
+        log.critical(e)
+        sys.exit(1)
+    # Throw error and exit if the type of config file data is incorrect
+    if type(fileData) is not dict:
+        log.critical('ERROR: Config file data is not dict (object). Data '
+                     'must be formatted like: {"argname": "value"}')
+        sys.exit(1)
+    # Iterate the key, value items in the config file
+    for arg, value in fileData.items():
+        if arg == "command":  # If we are setting the command we are running
+            # Just set it now as it will not exist in the argparse namespace
+            #     if the type was not set at all
+            parsed_args.command = value
+        # If the argument is not an argument in the parser
+        if arg not in parsed_args.__dict__:
+            log.critical(f'ERROR: Argument from file ({arg}) is not a '
+                         'recognized argument for this program')
+            sys.exit(1)
+        # If the value was not set in the CLI
+        if not getattr(parsed_args, arg):
+            # If the argument is the API key and there is an environment var
+            if arg == 'apiKey' and os.environ.get('MERAKI_DASHBOARD_API_KEY'):
+                continue  # Don't set the value, prefer the env variable
+            # Set the attribute
+            setattr(parsed_args, arg, value)
+
+
+def _args_from_file(parsed_args: argparse.Namespace) -> None:
+    """
+    Check some pre-set and CLI-provided locations for a config file and load
+        config parameters from that file.
+
+    - parsed_args: Parsed arguments namespace object from argparse.
+    """
+    # Static file paths to check
+    filePaths = [
+        'meraki-cli.conf',  # Current working directory
+        os.path.join(  # Hidden dot (.) folder in user's directory
+            os.path.expanduser("~"),
+            '.meraki-cli',
+            'meraki-cli.conf'
+            ),
+        os.path.join(  # Standard MacOS app config directory
+            os.path.expanduser("~"),
+            'Library',
+            'Application Support',
+            'meraki-cli',
+            'meraki-cli.conf'
+            ),
+        '/etc/meraki-cli/meraki-cli.conf',  # Standard Linux config directory
+    ]
+    # Environment variable directories to check
+    envPaths = [
+        'APPDATA',  # Windows roamed user settings directory
+        'LOCALAPPDATA',  # Windows machine-specific user settings directory
+    ]
+    # If a config file path was provided from the CLI
+    if parsed_args.configFile:
+        # If that file doesn't exist, throw error and exit
+        if not os.path.isfile(parsed_args.configFile):
+            log.critical(f'Defined config file ({parsed_args.configFile})'
+                         ' does not exist')
+            sys.exit(1)
+        # If it does exist, process it
+        _reconcile_args(parsed_args, parsed_args.configFile)
+    else:  # If a path wasn't provided from the CLI
+        # Loop through environment variable paths
+        for envPath in envPaths:
+            # Grab the value (None if non-existant)
+            value = os.environ.get(envPath)
+            if value:  # If there was an env variable
+                # Use ./meraki/meraki.conf as the path
+                filePath = os.path.join(value, 'meraki-cli', 'meraki-cli.conf')
+                # If the file exists in that path
+                if os.path.isfile(filePath):
+                    # Process it
+                    _reconcile_args(parsed_args, filePath)
+                    return None
+        # Loop through the static file paths
+        for filePath in filePaths:
+            if os.path.isfile(filePath):  # If this file exists
+                _reconcile_args(parsed_args, filePath)  # Process it
+                return None  # And quit the function
+
+
 def _configure_logging(parsed_args: argparse.Namespace) -> ():
     """
     Prepare and return two logging objects: one for the general program, and
@@ -242,8 +359,20 @@ def _cmd_help(arg_obj: Args) -> str:
     - mtdstr: String name of the method (ie: 'getOrganizations')
     """
     result = arg_obj.method.__doc__  # Start with the doc string
+    # Set the docstring header (wrapped in asterisks) to uppercase
+    result = re.sub(r'\*\*(.*)\*\*', lambda ele: ele.group(0).upper(), result)
+    # Remove leading tabs from header and add en extra line break
+    result = re.sub(r'\t\t\*\*(.*)\*\*', r'\1\n', result)
+    # Remove leading tabs from link
+    result = re.sub(r'\t\t(https://.*)\n', r'\1', result)
+    if arg_obj.positionals:  # If there are positional (required) arguments
+        # Add an "All Arguments" line below http for the argument entries
+        result = re.sub(r'(https://.*)\n', r'\1\n\nAll Arguments:\n', result)
+    # Remove leading tabs from argument entries and append double-hypen
+    result = re.sub(r'\t\t- ([a-zA-Z0-9_]+) ', r'  --\1 ', result)
     # Add the signature at the bottom of the help.
-    result += f'\n        >>> def {arg_obj.name}{arg_obj.signature}:'
+    result += '\nFunction Signature:\n\t'\
+              f'>>> def {arg_obj.name}{arg_obj.signature}:'
     return result
 
 
@@ -257,9 +386,9 @@ def _clean_args(parsed_args: argparse.Namespace):
     # Start with a copy of the namespace dict so it can be modified.
     arg_dict = dict(parsed_args.__dict__)
     # Iterate a list of known static arguments
-    for key in ['apiKey', 'debug', 'logfile', 'jsonOutput', 'type',
-                'filters', 'and_logic', 'command', 'kwargs', 'columns',
-                'translations', 'output_commands']:
+    for key in ['apiKey', 'debug', 'logfile', 'jsonOutput', 'configFile',
+                'type', 'filters', 'and_logic', 'command', 'kwargs',
+                'columns', 'translations', 'output_commands']:
         try:
             del arg_dict[key]  # Remove each from the dict
         except KeyError:
@@ -408,6 +537,60 @@ def _get_stdin_args(parsed_args: argparse.Namespace, arg_obj: Args) -> []:
     return _json_to_args(stdin, parsed_args, arg_obj)
 
 
+class ExtrasParser(argparse.ArgumentParser):
+    """
+    Customised argparse class for minimal errors on parsing extra args for
+        use in kwargs
+    """
+    def error(self, message: str) -> None:
+        """
+        Override the error method to output minimal info
+        """
+        log.critical(message)  # Output using our logging
+        sys.exit(2)  # Exit with an error code
+
+
+def _extra_args_to_kwargs(existing_kwarg_dict: dict, arglist: list) -> {}:
+    """
+    Interpret a list of 'extra' uninterpreted arguments from the CLI into a
+        dict of arguments which are then combined with the existing kwargs
+        dict. This resulting kwargs dict is then passed into the target method
+        for keyword parameters as well as target_method(**kwargs) for POST
+        bodies in the API.
+
+    - existing_kwarg_dict: Dict of existing kwargs from the main interpreter
+    - arglist: List of extra, unrecognized CLI arguments discarded by the
+        primary parser.
+    """
+    if not arglist:  # If there were no extra arguments provided
+        return existing_kwarg_dict  # Just return what we started with
+    parser = ExtrasParser()  # Instantiate our custom parser class
+    for arg in arglist:  # For each of the words in the unparsed arguments
+        if arg[:2] == '--':  # If the args starts with two hyphens
+            parser.add_argument(arg)  # Add it as an argument in the parser
+    # Parse known args so we ignore any extra info in the CLI (like comments)
+    #     and grab the first entry from the tuple (since it will pass thru
+    #     any other extras)
+    args = parser.parse_known_args(arglist)[0]
+    extra_arg_dict = {}  # Start with an empty dict for the extra args
+    # Iterate each key, value pair in the parsed args
+    for key, value in args.__dict__.items():
+        try:  # Allow JSON parsing to fail
+            # Try to read as JSON data in case it is nested
+            data = json.loads(value)
+            # If the data is a list or a dict
+            if type(data) is list or type(data) is dict:
+                # Then add it as that native data type
+                extra_arg_dict[key] = json.loads(value)
+            else:  # If it is not a list or a dict
+                extra_arg_dict[key] = value  # Add it as the standard string
+        except Exception:  # If this was non-compliant JSON data
+            extra_arg_dict[key] = value  # Add it as the stnadard string
+    # Merge the extra_arg_dict into the existing kwargs
+    existing_kwarg_dict.update(extra_arg_dict)
+    return existing_kwarg_dict  # And return it for use in the target method
+
+
 def _print_commands(parsed_args: argparse.Namespace,
                     argtups: tuple, arg_obj: Args) -> None:
     """
@@ -511,13 +694,34 @@ def _table_ready_dicts(listofdicts: list) -> []:
         set of object atributes where the key will be the table column name.
         The dicts are prepared by discarding any values which cannot be easily
         printed into a table cell (like lists, dicts, etc).
+    The input should be a list of dictionaries, but might be something else.
+        If it is something else, then we need to modify it to be ready for
+        tabluation.
 
     - listofdicts: List of dictionary objects where each dictionary is a series
         of attributes which can be matched against and filtered. Example:
         [{'id': '1', 'name': 'THING1'}, {'id': '2', 'name': 'THING2'}]
     """
-    bad_types = [dict, list]  # Types of data we want to discard
     result = []
+    #################################################
+    # Check For Nested Data
+    if type(listofdicts) is dict:  # If we were given a dict instead of a list
+        if (
+                len(listofdicts) == 1  # If the dict has one kv pair
+                and
+                # The value of that kv pair is a list
+                type(list(listofdicts.values())[0]) is list
+                and
+                # The entries in that list are dicts
+                type(list(listofdicts.values())[0][0]) is dict
+                ):
+            # Use the lsit from the kv pair as the list of dicts
+            listofdicts = list(listofdicts.values())[0]
+        else:
+            listofdicts = [listofdicts]
+    #################################################
+    # Strip Bad Types
+    bad_types = [dict, list]  # Types of data we want to discard
     for d in listofdicts:
         newdict = {}  # Use a new dict instead of deleting kv pairs
         for k, v in d.items():
@@ -526,6 +730,7 @@ def _table_ready_dicts(listofdicts: list) -> []:
             else:
                 log.debug(f'Discarding data in table squash: {k}: {v}')
         result.append(newdict)
+    #################################################
     return result
 
 
@@ -601,9 +806,6 @@ def _output(parsed_args: argparse.Namespace, result: list) -> None:
                   f'Forward pipelining: {not NO_STDOUT}')
         print(json.dumps(result, indent=4))
     else:  # Otherwise let's try to output a nice table.
-        # Make data a list if it is not already
-        if type(result) is dict:
-            result = [result]
         # Prepare data to be formatted into a table by removing value
         #     types unfriendly to tabulation.
         tabledicts = _table_ready_dicts(result)
@@ -622,7 +824,7 @@ def main(argstring=None) -> None:
     #     argparse structure from it
     api = meraki.DashboardAPI('fake_key', suppress_logging=True)
     # Start our arg parser instance
-    parser = argparse.ArgumentParser(add_help=False)
+    parser = Parser(add_help=False)
     argcomplete.autocomplete(parser)
     # Structure top level arguments into groups to make them easier to read.
     #     Any arguments added here need to also be listed in the _clean_args
@@ -649,12 +851,16 @@ def main(argstring=None) -> None:
                              help='Write logs to a logfile',
                              metavar='PATH',
                              dest='logfile')
+    basic_group.add_argument('-c', '--configFile',
+                             help='Use a config file for some arguments',
+                             metavar='PATH',
+                             dest='configFile')
     output_group = parser.add_argument_group('Output Formatting')
     output_group.add_argument('-j', '--jsonOutput',
                               help='Format output as JSON insted of default',
                               dest='jsonOutput',
                               action='store_true')
-    output_group.add_argument('-c', '--columns',
+    output_group.add_argument('-s', '--columns',
                               help='Filter/Order Table Columns '
                               '(ie: id,networkId,etc)',
                               metavar='STRING',
@@ -713,49 +919,62 @@ def main(argstring=None) -> None:
             #     options
             cmd = tsub.add_parser(
                 mtdstr,  # Use the method name for the command endpoint
+                add_help=False,  # Disable auto help since we add it manually
                 help=_cmd_title(mtdstr),  # Grab the help section (title)
                 description=_cmd_help(arg_obj),  # Grab the help section
                 # Use a raw formatter to allow the help docstring to maintain
                 #     it's multi-line format
                 formatter_class=argparse.RawTextHelpFormatter)
+            # A nice group for required (positional) arguments
+            req_group = cmd.add_argument_group('Required Arguments')
+            # A nice group for known optional (kwargs, help) arguments
+            msc_group = cmd.add_argument_group('Misc Arguments')
+            # Add help manually
+            msc_group.add_argument('-h', '--help',
+                                   help='Show help for this command',
+                                   action='help')
             # Add the method parameter info into the map
             parser_map[clsstr][mtdstr] = arg_obj
             # Iterate the positionals and create argument options for each
             for arg in arg_obj.positionals:
-                cmd.add_argument(f'--{arg.name}',
-                                 dest=arg.name,
-                                 help='(required)',
-                                 # Don't require the arg if we are
-                                 #     parsing the STDIN data
-                                 required=NO_STDIN,
-                                 # Use the annotation map to pass
-                                 #     additional params to the
-                                 #     argument option to make it
-                                 #     easier to use and read
-                                 **ANNOTATION_MAP[arg.annotation])
+                req_group.add_argument(f'--{arg.name}',
+                                       dest=arg.name,
+                                       help='(required)',
+                                       # Don't require the arg if we are
+                                       #     parsing the STDIN data
+                                       required=NO_STDIN,
+                                       # Use the annotation map to pass
+                                       #     additional params to the
+                                       #     argument option to make it
+                                       #     easier to use and read
+                                       **ANNOTATION_MAP[arg.annotation])
                 # Add the parameter name to the positionals list in the map
             for arg in arg_obj.keywords:
-                cmd.add_argument(f'--{arg.name}',
-                                 dest=arg.name,
-                                 default=arg.default,
-                                 # NOTE: These are never required.
-                                 help=f'(default = {arg.default})',
-                                 **ANNOTATION_MAP[type(arg.default)])
+                msc_group.add_argument(f'--{arg.name}',
+                                       dest=arg.name,
+                                       default=arg.default,
+                                       # NOTE: These are never required.
+                                       help=f'(default = {arg.default})',
+                                       **ANNOTATION_MAP[type(arg.default)])
             if arg_obj.varkw:
-                cmd.add_argument(f'--{arg_obj.varkw.name}',
-                                 dest=arg_obj.varkw.name,
-                                 help='(JSON formatted extra arguments)',
-                                 metavar='JSON_STRING',)
+                msc_group.add_argument(f'--{arg_obj.varkw.name}',
+                                       dest=arg_obj.varkw.name,
+                                       help='(Advanced Users) Optional '
+                                            'arguments in JSON format ',
+                                       metavar='JSON_STRING',)
     # If an argstring was passed in, we are probably being tested
     if argstring:
         # Split up the args and pass them in as a list to be parsed
-        args = parser.parse_args(argstring.split(' '))
+        args, extraarglist = parser.parse_known_args(argstring.split(' '))
     else:
-        args = parser.parse_args()  # Parse the user provided CLI commands/args
+        # Parse the user provided CLI commands/args
+        args, extraarglist = parser.parse_known_args()
     global log  # Make the "log" variable global to anybody can use it
+    log.setLevel(logging.DEBUG)
+    _args_from_file(args)  # Process any arguments from a file
     # Pull in the two logging systems
     log, meraki_log = _configure_logging(args)
-    log.debug(f'CLI Arguments: {args}')
+    log.debug(f'Argument Settings: \n{json.dumps(args.__dict__, indent=4)}')
     if not args.type:  # If a type (networks, switch, etc) wasn't specified
         parser.print_help()  # Print help so the user can see options
         sys.exit()  # And exit the program
@@ -822,6 +1041,8 @@ def main(argstring=None) -> None:
             arg_dict = argtup[1]
         else:
             arg_dict = {}
+        if extraarglist:
+            arg_dict = _extra_args_to_kwargs(arg_dict, extraarglist)
         log.info('Calling target method with:')
         log.info(f'    *Positionals: {positionals}')
         log.info(f'    **Named/Kwargs: {arg_dict}')
